#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <jni.h>
#include <android/log.h>

typedef unsigned long UINT_PTR;

#include "CPP/myWindows/StdAfx.h"
#include "CPP/Common/MyWindows.h"
#include "CPP/Common/IntToString.h"
#include "CPP/Common/MyInitGuid.h"
#include "CPP/Common/StringConvert.h"
#include "CPP/Common/ComTry.h"

#include "CPP/Windows/DLL.h"
#include "CPP/Windows/FileDir.h"
#include "CPP/Windows/FileFind.h"
#include "CPP/Windows/FileName.h"
#include "CPP/Windows/NtCheck.h"
#include "CPP/Windows/PropVariant.h"
#include "CPP/Windows/COM.h"

#include "CPP/7zip/Common/FileStreams.h"
#include "CPP/7zip/Archive/IArchive.h"

#include "CPP/7zip/ICoder.h"
#include "CPP/7zip/IPassword.h"
#include "CPP/7zip/MyVersion.h"
#include "CPP/7zip/Common/CreateCoder.h"
#include "CPP/7zip/Archive/IArchive.h"

#include "C/cvtutf.h"
#include "CPP/Windows/Synchronization.h"

#define COUNTOF(a) (sizeof(a)/sizeof(a[0]))

#define EXTRACT_EXTRACT 0
#define EXTRACT_SKIP 1
#define EXTRACT_ABORT 2
#define EXTRACT_TOMEM 3

#define UN7ZIPERROR_UNSUPPORTED 1
#define UN7ZIPERROR_CRCERROR 2
#define UN7ZIPERROR_DATAERROR 4
#define UN7ZIPERROR_ABORTED 8
#define UN7ZIPERROR_CREATE 16
#define UN7ZIPERROR_DELETE 32
#define UN7ZIPERROR_MEMORY 64

#define _UNIX
#define SUCCEEDED(hr) (((HRESULT)(hr)) >= 0)

namespace NArchive {
    namespace N7z {

        using namespace NWindows;

        JavaVM *gJavaVM = NULL;

        static NWindows::NSynchronization::CCriticalSection g_FileCountCriticalSection;

#define DEFINE_GUID2(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
    MY_EXTERN_C const GUID name = { l, w1, w2, { b1, b2, b3, b4, b5, b6, b7, b8 } }

        void itow(int n, WCHAR *Str) {
            WCHAR NumStr[50];
            size_t Pos = 0;

            do {
                NumStr[Pos++] = (((char) (n)) % 10) + '0';
                n = n / 10;
            } while (n != 0);

            for (size_t I = 0; I < Pos; I++)
                Str[I] = NumStr[Pos - I - 1];
            Str[Pos] = 0;
        }

        jlong atoilw(const WCHAR *s) {
            bool sign = false;
            if (*s == '-') // We do use signed integers here, for example, in GUI SFX.
            {
                s++;
                sign = true;
            }
            // Use unsigned type here, since long string can overflow the variable
            // and signed integer overflow is undefined behavior in C++.
            jlong n = 0;
            while (*s >= '0' && *s <= '9') {
                n = n * 10 + (*s - '0');
                s++;
            }
            // Check int64(n)>=0 to avoid the signed overflow with undefined behavior
            // when negating 0x8000000000000000.
            return sign && jlong(n) >= 0 ? -jlong(n) : jlong(n);
        }

        int atoiw(const WCHAR *s) {
            return (int) atoilw(s);
        }

        wchar_t *mywcscpy(wchar_t *d, const wchar_t *s) {
            wchar_t *a = d;
            while ((*d++ = *s++));
            return a;
        }

        wchar_t *mywcsncpy(wchar_t *d, const wchar_t *s, int maxlen) {
            wchar_t *a = d;
            while ((*d++ = *s++) && maxlen-- > 0);
            return a;
        }

        int mywcsncmp(const wchar_t *l, const wchar_t *r, int maxlen)
        {
            for (; *l==*r && *l && *r && maxlen>=0; maxlen--, l++, r++);
            if (maxlen==0)
                return 0;
            return *l - *r;
        }

        wchar_t *mywcslcat(wchar_t *dest, const wchar_t *src,
                           int maxlen)  // maxlen is the max of the combined string!
        {
            mywcsncpy(dest + mywcslen(dest), src, maxlen - mywcslen(dest));
            dest[maxlen] = 0;
            return dest;
        }

        WCHAR *mywcslcatslash(WCHAR *thedir, int maxlen) {
            if (thedir[0] && mywcslen(thedir) < (DWORD) maxlen)
                if (thedir[mywcslen(thedir) - 1] != '/') {
                    mywcslcat(thedir, L"/", maxlen);
                    thedir[maxlen] = 0;
                }
            return thedir;
        }

        WCHAR *mywcslcpy(WCHAR *p, const WCHAR *p2, int maxlen) {
            if ((int) mywcslen(p2) >= maxlen) {
                mywcsncpy(p, p2, maxlen);
                p[maxlen] = 0;
            } else
                mywcscpy(p, p2);
            return p;
        }

        WCHAR *mywcschr(WCHAR *p, const WCHAR c) {
            WCHAR *ret = NULL;
            do {
                if (*p == c) {
                    ret = p;
                    break;
                }
            } while (*p++ != 0);
            return ret;
        }

        WCHAR *mywcsrchr(WCHAR *p, const WCHAR c) {
            WCHAR *ret = NULL;
            do {
                if (*p == c)
                    ret = p;
            } while (*p++ != 0);
            return ret;
        }

        BOOL buf2ToHex(WCHAR* buf,WCHAR* c) {
            if (buf[0]>='a' && buf[0]<='f')
                buf[0]=buf[0]+'A'-'a';
            if (buf[1]>='a' && buf[1]<='f')
                buf[1]=buf[1]+'A'-'a';
            int v1,v2;
            if (buf[0]>='0' && buf[0]<='9')
                v1=buf[0]-'0';
            else if (buf[0]>='A' && buf[0]<='F')
                v1=10+buf[0]-'A';
            else
                return false;
            if (buf[1]>='0' && buf[1]<='9')
                v2=buf[1]-'0';
            else if (buf[1]>='A' && buf[1]<='F')
                v2=10+buf[1]-'A';
            else
                return false;
            *c=(WCHAR)(v1*16+v2);
            return true;
        }

        void decodeUrl(WCHAR *out,const WCHAR* in,int maxlen) {
            WCHAR *outend=out+maxlen;
            WCHAR c = 'x';
            while (c) {
                if (out+1>=outend) {
                    out++[0]=0;
                    break;
                }
                c = in++[0];
                switch (c) {
                    case '+':
                        out++[0] = ' ';
                    case '%':
                        if (in[1] != 0) {
                            WCHAR buf[3];
                            buf[0] = in[0];
                            buf[1] = in[1];
                            buf[2] = 0;
                            if (buf2ToHex(buf, &c)) {
                                out++[0] = c;
                                in+=2;
                            } else
                                out++[0] = '%';
                        } else
                            out++[0] = '%';
                    default:
                        out++[0] = c;
                }
            }
        }

        DEFINE_GUID2(CLSID_CFormatbzip2,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x02, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatrar,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x03, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatarj,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x04, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatz,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x05, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatlzh,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x06, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormat7z2,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x07, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatcab,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x08, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatxz,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0x0C, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatrar5,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0xCC, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatiso,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0xE7, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormattar,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0xEE, 0x00, 0x00);
        DEFINE_GUID2(CLSID_CFormatgz,
                     0x23170F69, 0x40C1, 0x278A, 0x10, 0x00, 0x00, 0x01, 0x10, 0xEF, 0x00, 0x00);

        DEFINE_GUID2(IID_IInArchive2,
                     0x23170F69, 0x40C1, 0x278A, 0x00, 0x00, 0x00, 0x06, 0x00, 0x60, 0x00, 0x00);

        DEFINE_GUID(CLSID_CArchiveHandler,
                    k_7zip_GUID_Data1,
                    k_7zip_GUID_Data2,
                    k_7zip_GUID_Data3_Common,
                    0x10, 0x00, 0x00, 0x01, 0x10, 0x00, 0x00, 0x00);

        STDAPI CreateArchiver(const GUID *clsid, const GUID *iid, void **outObject);

        jmethodID openTargetDocumentGlobal;
        jmethodID openSourceDocumentGlobal;
        jmethodID closeDocumentGlobal;
        jmethodID writeTargetDocumentGlobal;
        jmethodID getFdGlobal;
        int CALLBACKBUFSIZE = 1024 * 1024;
        jbyteArray callbackbuf;
        jobject globalStreamCallback = NULL;
        volatile int openFileCount = 0;

        class CMyInFileStream :
                public CInFileStream {
        public:
            UInt64 FirstVolumeSize, CurrentVolumeOffset, LastOffsetInMultiVolume;
            int CurrentVolumeNr; // 1-based!
            WCHAR VolumeBaseName[1024];
            WCHAR ArchiveBaseName[1024];
            bool isContentFile;
            int contentFileHandle, contentFileFd;

            bool Open(LPCWSTR fileName);

            virtual ~CMyInFileStream();

            STDMETHOD(Read)(void *data, UInt32 size, UInt32 *processedSize);

            STDMETHOD(Seek)(Int64 offset, UInt32 seekOrigin, UInt64 *newPosition);

        private:
            bool openContentStream(LPCWSTR fileName);

            bool closeContentStream();
        };

        bool CMyInFileStream::openContentStream(LPCWSTR fileName) {
            JNIEnv *envLocal;
            int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);
            jmethodID openSourceDocumentLocal;
            jclass callbackClass = NULL;
            if (status == JNI_EDETACHED) {   // called from different thread?
                status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                callbackClass = envLocal->GetObjectClass(globalStreamCallback);
                openSourceDocumentLocal = envLocal->GetMethodID(callbackClass, "openSourceDocument",
                                                                "(Ljava/lang/String;)I");
            } else {
                status = -12345;
                openSourceDocumentLocal = openSourceDocumentGlobal;
            }
            jstring jfileName = envLocal->NewStringUTF(GetAnsiString(fileName));
            contentFileHandle = envLocal->CallIntMethod(globalStreamCallback,
                                                        openSourceDocumentLocal, jfileName);
            envLocal->DeleteLocalRef(jfileName);
            jmethodID getFdLocal;
            if (contentFileHandle != 0) {
                if (status == 0)
                    getFdLocal = envLocal->GetMethodID(callbackClass, "getFd", "(I)I");
                else
                    getFdLocal = getFdGlobal;
                contentFileFd = envLocal->CallIntMethod(globalStreamCallback, getFdLocal,
                                                        contentFileHandle);
            }
            if (status == 0)
                gJavaVM->DetachCurrentThread();
            return contentFileHandle != 0;
        }

        bool CMyInFileStream::Open(LPCWSTR fileName) {
            FirstVolumeSize = 0;
            CurrentVolumeOffset = 0;
            LastOffsetInMultiVolume = 0;
            CurrentVolumeNr = 0;
            ArchiveBaseName[0] = 0;
            VolumeBaseName[0] = 0;
            mywcslcat(VolumeBaseName, fileName, 1023);
            isContentFile = (fileName[0] == 'c') ||
                            (fileName[0] == '/' && fileName[1] == '/' && fileName[2] == '/' &&
                             fileName[3] == '_');
            if (isContentFile) {
                return openContentStream(fileName);
            } else
                return CInFileStream::Open(fileName);
        }

        bool CMyInFileStream::closeContentStream() {
            if (isContentFile && contentFileHandle != 0) {
                JNIEnv *envLocal;
                int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);
                jmethodID closeDocumentLocal;
                if (status == JNI_EDETACHED) {   // called from different thread?
                    status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                    jclass callbackClass = envLocal->GetObjectClass(globalStreamCallback);
                    closeDocumentLocal = envLocal->GetMethodID(callbackClass, "closeDocument",
                                                               "(I)V");
                } else {
                    status = -12345;
                    closeDocumentLocal = closeDocumentGlobal;
                }
                envLocal->CallVoidMethod(globalStreamCallback, closeDocumentLocal,
                                         contentFileHandle);
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
                contentFileHandle = 0;
                return true;
            }
            return false;
        }

        CMyInFileStream::~CMyInFileStream() {
            closeContentStream();
        }

        STDMETHODIMP CMyInFileStream::Read(void *data, UInt32 size, UInt32 *processedSize) {
            UInt32 len, len2;
            UInt64 newPosition, seekpos;
            HRESULT ret;
            len = 0;
            if (isContentFile) {
                if (size == 0) {
                    if (processedSize)
                        *processedSize = 0;
                    return S_OK;
                }
                int len3;
                do {
                    len3 = read(contentFileFd, data, size);
                } while (len3 < 0 && (errno == EINTR));
                len = (UInt32) (len3);
                ret = (len3 >= 0) ? S_OK : E_FAIL;
            } else
                ret = CInFileStream::Read(data, size, &len);

            if (FirstVolumeSize > 0 && SUCCEEDED(ret)) {   // Multi-part archive!
                LastOffsetInMultiVolume = LastOffsetInMultiVolume + len;
                while (len < size && SUCCEEDED(ret)) {   // Continued in next volume?
                    newPosition = 0;
                    seekpos = LastOffsetInMultiVolume;
                    Seek(seekpos, 0, &newPosition);
                    if (newPosition == seekpos) {  // Try to read the rest!
                        len2 = 0;
                        if (isContentFile) {
                            int len3;
                            do {
                                len3 = read(contentFileFd, ((char *) data) + len, size - len);
                            } while (len3 < 0 && (errno == EINTR));
                            len2 = (UInt32) len3;
                            ret = (len3 >= 0) ? S_OK : E_FAIL;
                        } else
                            ret = CInFileStream::Read(((char *) data) + len, size - len, &len2);
                        if (len2 <= 0)
                            break;

                        len += len2;
                        LastOffsetInMultiVolume += len2;
                    } else
                        break;
                }
            }
            if (processedSize)
                *processedSize = len;
            return ret;
        }

        STDMETHODIMP CMyInFileStream::Seek(Int64 offset, UInt32 seekOrigin, UInt64 *newPosition) {
            UInt64 newpos, neededpos;
            int neededvolume;
            WCHAR namebuf[1024];
            WCHAR buf2[32];
            int i;
            if (FirstVolumeSize > 0) {   //Multi-part archive!
                if (seekOrigin == 0)
                    neededpos = offset;
                else
                    neededpos = LastOffsetInMultiVolume + offset;
                neededvolume = 1 + (int) (neededpos / FirstVolumeSize);
                if (neededvolume != CurrentVolumeNr) {
                    namebuf[0] = 0;
                    mywcslcat(namebuf, VolumeBaseName, 1023);
                    itow(neededvolume, buf2);
                    for (i = mywcslen(buf2); i <= 2; i++)
                        mywcslcat(namebuf, L"0", 1023);
                    mywcslcat(namebuf, buf2, 1023);
                    if (isContentFile) {
                        closeContentStream();
                        if (!openContentStream(namebuf)) {
                            if (newPosition)
                                *newPosition = LastOffsetInMultiVolume;
                            return E_FAIL;
                        }
                    } else {
                        File.Close();
                        if (!SUCCEEDED(CInFileStream::Open(namebuf))) {
                            if (newPosition)
                                *newPosition = LastOffsetInMultiVolume;
                            return E_FAIL;
                        }
                    }
                    CurrentVolumeNr = neededvolume;
                    CurrentVolumeOffset = (CurrentVolumeNr - 1) * FirstVolumeSize;
                }
                HRESULT ret;
                if (isContentFile) {
                    newpos = lseek64(contentFileFd, neededpos - CurrentVolumeOffset, 0);
                    ret = newpos == -1 ? E_FAIL : S_OK;
                } else
                    ret = CInFileStream::Seek(neededpos - CurrentVolumeOffset, 0, &newpos);
                LastOffsetInMultiVolume = newpos + CurrentVolumeOffset;
                if (newPosition)
                    *newPosition = LastOffsetInMultiVolume;
                return ret;
            } else {
                if (isContentFile) {
                    Int64 newPosition2 = lseek64(contentFileFd, offset, seekOrigin);
                    if (newPosition)
                        *newPosition = (UInt64) newPosition2;
                    return newPosition2 == -1 ? E_FAIL : S_OK;
                } else
                    return CInFileStream::Seek(offset, seekOrigin, newPosition);
            }
        }

        class CArchiveOpenCallback :
                public IArchiveOpenCallback,
                public IArchiveOpenVolumeCallback,
                public ICryptoGetTextPassword,
                public CMyUnknownImp {
        public:
            MY_UNKNOWN_IMP2(ICryptoGetTextPassword,IArchiveOpenVolumeCallback)

            STDMETHOD(SetTotal)(const UInt64 *files, const UInt64 *bytes);

            STDMETHOD(SetCompleted)(const UInt64 *files, const UInt64 *bytes);

            STDMETHOD(CryptoGetTextPassword)(BSTR *password);

            STDMETHOD(GetProperty)(PROPID propID, PROPVARIANT *value);

            STDMETHOD(GetStream)(const wchar_t *name, IInStream **inStream);

            void Init(JNIEnv *env1, jobject callback1, const WCHAR* archiveBaseName);
            JNIEnv *env;
            jobject callback;
            jmethodID onGetPassword;
            void *_userData;
//    tPasswordCallback PasswordCallback;
        private:
            WCHAR _archiveBaseName[1024];// archive file name
        };


        STDMETHODIMP
        CArchiveOpenCallback::SetTotal(const UInt64 * /* files */, const UInt64 * /* bytes */) {
            return S_OK;
        }

        STDMETHODIMP
        CArchiveOpenCallback::SetCompleted(const UInt64 * /* files */, const UInt64 * /* bytes */) {
            return S_OK;
        }

        STDMETHODIMP CArchiveOpenCallback::CryptoGetTextPassword(BSTR *password) {
            if (onGetPassword != 0) {
                __android_log_write(ANDROID_LOG_VERBOSE, "Un7zip", "onGetPassword");
                jobject pw = env->CallObjectMethod(callback, onGetPassword);
                if (pw != 0) {
                    const char *pwstr = env->GetStringUTFChars((jstring) pw, 0);
                    if (pwstr != NULL) {
                        int ret = StringToBstr(GetUnicodeString(pwstr), password);
                        env->ReleaseStringUTFChars((jstring) pw, pwstr);
                        return S_OK;
                    }
                    env->DeleteLocalRef(pw);
                }
            }
            return E_ABORT;
        }

        void CArchiveOpenCallback::Init(JNIEnv *env1, jobject callback1, const WCHAR* archiveBaseName) {
            env = env1;
            callback = callback1;
            mywcslcpy(_archiveBaseName, archiveBaseName, 1023);
            if (callback1 != NULL) {
                jclass callbackClass = env->GetObjectClass(callback);
                onGetPassword = env->GetMethodID(callbackClass, "onGetPassword","()Ljava/lang/String;");
            }
        }

        STDMETHODIMP CArchiveOpenCallback::GetProperty(PROPID propID, PROPVARIANT *value) {
            if ( propID == kpidName )
            {
                value->vt = VT_BSTR;
                value->bstrVal = SysAllocString(_archiveBaseName);
            }

            if ( propID == kpidSize )
            {
                value->vt = VT_UI8;
                value->uhVal.QuadPart = 0;
                return E_FAIL;
            }

            return S_OK;
        }

        STDMETHODIMP CArchiveOpenCallback::GetStream(const wchar_t *name, IInStream **inStream) {
            CMyInFileStream *fileSpec = new CMyInFileStream;
            CMyComPtr<IInStream> file = fileSpec;

            if (!fileSpec->Open(name)) {
                return E_FAIL;
            }
            *inStream=file.Detach();
            return 0;
        }

        class COutDocStream :            // using documents API!
                public IOutStream,
                public CMyUnknownImp {
            UInt64 fixedmempos;
            UInt64 fixedmemdatalen;
            int fileHandle;

        public:
            virtual ~COutDocStream() {}

            bool Open(WCHAR *fileName) {
                fixedmempos = 0;
                fixedmemdatalen = 0;
                JNIEnv *envLocal;
                int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);
                jmethodID openTargetDocumentLocal;
                if (status == JNI_EDETACHED) {   // called from different thread?
                    status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                    jclass callbackClass = envLocal->GetObjectClass(globalStreamCallback);
                    openTargetDocumentLocal = envLocal->GetMethodID(callbackClass,
                                                                    "openTargetDocument",
                                                                    "(Ljava/lang/String;)I");
                } else {
                    status = -12345;
                    openTargetDocumentLocal = openTargetDocumentGlobal;
                }
                jstring jfileName = envLocal->NewStringUTF(GetAnsiString(fileName));
                fileHandle = envLocal->CallIntMethod(globalStreamCallback, openTargetDocumentLocal,
                                                     jfileName);
                if (fileHandle != 0) {
                    g_FileCountCriticalSection.Enter();
                    openFileCount++;
                    g_FileCountCriticalSection.Leave();
                }
                envLocal->DeleteLocalRef(jfileName);
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
                return fileHandle != 0;
            }

            HRESULT Close() {
                if (fileHandle) {
                    JNIEnv *envLocal;
                    g_FileCountCriticalSection.Enter();
                    openFileCount--;
                    g_FileCountCriticalSection.Leave();
                    int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);
                    jmethodID closeDocumentLocal;
                    if (status == JNI_EDETACHED) {   // called from different thread?
                        status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                        jclass callbackClass = envLocal->GetObjectClass(globalStreamCallback);
                        closeDocumentLocal = envLocal->GetMethodID(callbackClass, "closeDocument",
                                                                   "(I)V");
                    } else {
                        status = -12345;
                        closeDocumentLocal = closeDocumentGlobal;
                    }
                    envLocal->CallVoidMethod(globalStreamCallback, closeDocumentLocal, fileHandle);
                    if (status == 0)
                        gJavaVM->DetachCurrentThread();
                    fileHandle = 0;
                }
                return 0;
            }

            UInt64 ProcessedSize;

            bool SetTime(const FILETIME *cTime, const FILETIME *aTime, const FILETIME *mTime) {
                return true;
            }

            MY_UNKNOWN_IMP1(IOutStream)

            STDMETHOD(Write)(const void *data, UInt32 size, UInt32 *processedSize) {
                int maxread = size;
                int offset = 0;
                int ret = S_OK;
                if (maxread > 0) {
                    JNIEnv *envLocal;
                    int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);
                    jmethodID writeTargetDocumentLocal;
                    if (status == JNI_EDETACHED) {   // called from different thread?
                        status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                        jclass callbackClass = envLocal->GetObjectClass(globalStreamCallback);
                        writeTargetDocumentLocal = envLocal->GetMethodID(callbackClass,
                                                                         "writeTargetDocument",
                                                                         "(I[BI)I");
                    } else {
                        status = -12345;
                        writeTargetDocumentLocal = writeTargetDocumentGlobal;
                    }
                    if (callbackbuf == NULL) {
                        jbyteArray temp_buffer = envLocal->NewByteArray(CALLBACKBUFSIZE);
                        callbackbuf = (jbyteArray) envLocal->NewGlobalRef(temp_buffer);
                    }
                    int remain = maxread;
                    while (remain > 0) {
                        int partLen = remain;
                        if (partLen > CALLBACKBUFSIZE)
                            partLen = CALLBACKBUFSIZE;

                        g_FileCountCriticalSection.Enter();  // the memory buffer is shared between threads!
                        envLocal->SetByteArrayRegion(callbackbuf, 0, partLen,
                                                     (jbyte *) (data) + offset);
                        partLen = envLocal->CallIntMethod(globalStreamCallback,
                                                          writeTargetDocumentLocal,
                                                          fileHandle, callbackbuf, partLen);
                        g_FileCountCriticalSection.Leave();
                        if (partLen < 0) {
                            ret = E_FAIL;
                            break;
                        }
                        fixedmempos += partLen;
                        fixedmemdatalen = fixedmempos;
                        offset += partLen;
                        remain -= partLen;
                    }
                    if (status == 0)
                        gJavaVM->DetachCurrentThread();
                }
                if (processedSize)
                    *processedSize = (UInt32) maxread;
                return ret;
            }

            STDMETHOD(Seek)(Int64 offset, UInt32 seekOrigin, UInt64 *newPosition) {
                return E_FAIL;
            }

            STDMETHOD(SetSize)(UInt64 newSize) {
                fixedmemdatalen = (int) newSize;
                return S_OK;
            }
        };

        class COutTestOnlyStream :
                public IOutStream,
                public CMyUnknownImp {
            UInt64 fixedmempos;
            UInt64 fixedmemdatalen;

        public:
            virtual ~COutTestOnlyStream() {}

            bool Create() {
                return true;
            }

            bool Open() {
                fixedmempos = 0;
                fixedmemdatalen = 0;
                return true;
            }

            HRESULT Close();

            UInt64 ProcessedSize;

            bool SetTime(const FILETIME *cTime, const FILETIME *aTime, const FILETIME *mTime) {
                return true;
            }

            MY_UNKNOWN_IMP1(IOutStream)

            STDMETHOD(Write)(const void *data, UInt32 size, UInt32 *processedSize) {
                int maxread = size;
                if (maxread > 0) {
                    fixedmempos += maxread;
                    fixedmemdatalen = fixedmempos;
                }
                if (processedSize)
                    *processedSize = (UInt32) maxread;
                return S_OK;
            }

            STDMETHOD(Seek)(Int64 offset, UInt32 seekOrigin, UInt64 *newPosition) {
                Int64 newpos = fixedmempos;
                switch (seekOrigin) {
                    case STREAM_SEEK_SET:
                        newpos = offset;
                        break;
                    case STREAM_SEEK_CUR:
                        newpos = offset + fixedmempos;
                        break;
                    case STREAM_SEEK_END:
                        newpos = fixedmemdatalen + offset;
                        break;
                }
                if (newpos < 0)
                    newpos = 0;
                else if (newpos > fixedmemdatalen)
                    newpos = fixedmemdatalen;
                fixedmempos = (int) newpos;
                if (newPosition)
                    *newPosition = newpos;
                return S_OK;
            }

            STDMETHOD(SetSize)(UInt64 newSize) {
                fixedmemdatalen = (int) newSize;
                return S_OK;
            }
        };


        class COutFixedMemStream :
                public IOutStream,
                public CMyUnknownImp {
            char *fixedmem;
            int fixedmemsize, fixedmempos, fixedmemdatalen;
            int *datareadpointer;

        public:
            virtual ~COutFixedMemStream() {}

            bool Create(char *membuf, int membufsize, int *datareadptr) {
                datareadpointer = datareadptr;
                fixedmem = membuf;
                fixedmemsize = membufsize;
                fixedmempos = 0;
                fixedmemdatalen = 0;
                return membuf != NULL;
            }

            bool Open(LPCTSTR fileName, DWORD creationDisposition) {
                fixedmempos = 0;
                return fixedmem != NULL;
            }

            HRESULT Close();

            UInt64 ProcessedSize;

            bool SetTime(const FILETIME *cTime, const FILETIME *aTime, const FILETIME *mTime) {
                return true;
            }

            MY_UNKNOWN_IMP1(IOutStream)

            STDMETHOD(Write)(const void *data, UInt32 size, UInt32 *processedSize) {
                int maxread;
                if ((int) size > fixedmemsize - fixedmempos)
                    maxread = fixedmemsize - fixedmempos;
                else
                    maxread = size;
                if (maxread > 0) {
                    memcpy(fixedmem + fixedmempos, data, maxread);
                    fixedmempos += maxread;
                    if (fixedmempos > fixedmemdatalen) {
                        fixedmemdatalen = fixedmempos;
                        if (datareadpointer)
                            *datareadpointer = fixedmemdatalen;
                    }
                }
                if (processedSize)
                    *processedSize = (UInt32) maxread;
                return S_OK;
            }

            STDMETHOD(Seek)(Int64 offset, UInt32 seekOrigin, UInt64 *newPosition) {
                Int64 newpos = fixedmempos;
                switch (seekOrigin) {
                    case STREAM_SEEK_SET:
                        newpos = offset;
                        break;
                    case STREAM_SEEK_CUR:
                        newpos = offset + fixedmempos;
                        break;
                    case STREAM_SEEK_END:
                        newpos = fixedmemdatalen + offset;
                        break;
                }
                if (newpos < 0)
                    newpos = 0;
                else if (newpos > fixedmemdatalen)
                    newpos = fixedmemdatalen;
                fixedmempos = (int) newpos;
                if (newPosition)
                    *newPosition = newpos;
                return S_OK;
            }

            STDMETHOD(SetSize)(UInt64 newSize) {
                if (newSize <= fixedmemsize) {
                    fixedmemdatalen = (int) newSize;
                    return S_OK;
                }
                return E_FAIL;
            }
        };

//////////////////////////////////////////////////////////////
// Archive Extracting callback class

        static const wchar_t *kCantDeleteOutputFile = L"ERROR: Can not delete output file ";

        static const char *kTestingString = "Testing     ";
        static const char *kExtractingString = "Extracting  ";
        static const char *kSkippingString = "Skipping    ";

        static const char *kUnsupportedMethod = "Unsupported Method";
        static const char *kCRCFailed = "CRC Failed";
        static const char *kDataError = "Data Error";
        static const char *kUnknownError = "Unknown Error";

        class CArchiveExtractCallback :
                public IArchiveExtractCallback,
                public ICryptoGetTextPassword,
                public CMyUnknownImp {
        public:
            MY_UNKNOWN_IMP1(ICryptoGetTextPassword)

            // IProgress
            STDMETHOD(SetTotal)(UInt64 size);

            STDMETHOD(SetCompleted)(const UInt64 *completeValue);

            // IArchiveExtractCallback
            STDMETHOD(GetStream)(UInt32 index, ISequentialOutStream **outStream,
                                 Int32 askExtractMode);

            STDMETHOD(PrepareOperation)(Int32 askExtractMode);

            STDMETHOD(SetOperationResult)(Int32 resultEOperationResult);

            // ICryptoGetTextPassword
            STDMETHOD(CryptoGetTextPassword)(BSTR *aPassword);

        private:
            CMyComPtr<IInArchive> _archiveHandler;
            jlong _totalsize;
            WCHAR _directoryPath[1024];  // Output directory
            WCHAR _diskFilePath[1024];   // full path to file on disk
            bool _extractMode;
            bool _testOnly;
            jobject callback;
            struct CProcessedFileInfo {
                FILETIME MTime;
                UInt32 Attrib;
                bool isDir;
                bool AttribDefined;
                bool MTimeDefined;
            } _processedFileInfo;

            COutFileStream *_outFileStreamSpec;
            COutDocStream *_outDocStreamSpec;
            COutFixedMemStream *_outMemStreamSpec;
            COutTestOnlyStream *_outTestOnlyStreamSpec;
            CMyComPtr<ISequentialOutStream> _outFileStream;
        public:
            void Init(IInArchive *archiveHandler, const WCHAR *directoryPath, BOOL testOnly,
                      JNIEnv *env1, jobject callback1);

            UInt64 NumErrors;
            WCHAR _archiveBaseName[1024];// archive file name
            void *_userData;
            int subStart;  // cut start of path, or entire if -1
            int errorReason;
            UInt32 _lastsuccessfullycloseditem;
            UInt32 _lastopeneditem;
            JNIEnv *env;
            jmethodID onGetFileCount;
            jmethodID onUnpackProgress;
            jmethodID onWantExtract;
            jmethodID onGetPassword;
            jmethodID onCreateDir;
        };

        void CArchiveExtractCallback::Init(IInArchive *archiveHandler, const WCHAR *directoryPath,
                                           BOOL testOnly, JNIEnv *env1, jobject callback1) {
            NumErrors = 0;
            errorReason = 0;
            _archiveHandler = archiveHandler;
            _outFileStreamSpec = NULL;
            _outMemStreamSpec = NULL;
            _outTestOnlyStreamSpec = NULL;
            _outFileStream = NULL;
            _testOnly = testOnly;
            env = env1;
            callback = callback1;
            if (callback != NULL) {
                jclass callbackClass = env->GetObjectClass(callback);
                onGetFileCount = env->GetMethodID(callbackClass, "onGetFileCount", "(I)V");
                onUnpackProgress = env->GetMethodID(callbackClass, "onUnpackProgress", "(JJ)I");
                onWantExtract = env->GetMethodID(callbackClass, "onWantExtract",
                                                 "(ILjava/lang/String;Ljava/lang/String;ZZJJ)Ljava/lang/String;");
                onGetPassword = env->GetMethodID(callbackClass, "onGetPassword",
                                                 "()Ljava/lang/String;");
                onCreateDir = env->GetMethodID(callbackClass, "onCreateDir",
                                               "(Ljava/lang/String;)I");
            }

            if (directoryPath == NULL)
                _directoryPath[0] = 0;
            else {
                mywcslcpy(_directoryPath, directoryPath, 1023);
                mywcslcatslash(_directoryPath, 1023);
            }
        }

        STDMETHODIMP CArchiveExtractCallback::SetTotal(UInt64 size) {
            _totalsize = size;
            return S_OK;
        }

        STDMETHODIMP CArchiveExtractCallback::SetCompleted(const UInt64 *completeValue) {
            if (completeValue && onUnpackProgress) {
                JNIEnv *envLocal;
                int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);
                jmethodID onUnpackProgressLocal;
                if (status == JNI_EDETACHED) {   // called from different thread?
                    status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                    jclass callbackClass = envLocal->GetObjectClass(callback);
                    onUnpackProgressLocal = envLocal->GetMethodID(callbackClass, "onUnpackProgress",
                                                                  "(JJ)I");
                } else {
                    status = -12345;
                    onUnpackProgressLocal = onUnpackProgress;
                }
                BOOL ret = (envLocal->CallIntMethod(callback, onUnpackProgressLocal,
                                                    (jlong) *completeValue, (jlong) _totalsize) ==
                            0);
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
                if (!ret)
                    return E_ABORT;
            }
            return S_OK;
        }

        UInt64 ConvertPropVariantToUInt64(const PROPVARIANT &prop) {
            switch (prop.vt) {
                case VT_UI1:
                    return prop.bVal;
                case VT_UI2:
                    return prop.uiVal;
                case VT_UI4:
                    return prop.ulVal;
                case VT_UI8:
                    return (UInt64) prop.uhVal.QuadPart;
                case VT_BOOL:
                    return prop.boolVal ? 1 : 0;
                default:
                    return 0;
            }
        }

        HRESULT
        GetArchiveItemBoolProp(IInArchive *archive, UInt32 index, PROPID propID, bool &result) {
            NCOM::CPropVariant prop;
            result = false;
            RINOK(archive->GetProperty(index, propID, &prop));
            if (prop.vt == VT_BOOL)
                result = VARIANT_BOOLToBool(prop.boolVal);
            else if (prop.vt != VT_EMPTY)
                return E_FAIL;
            return S_OK;
        }

        HRESULT IsArchiveItemFolder(IInArchive *archive, UInt32 index, bool &result) {
            return GetArchiveItemBoolProp(archive, index, kpidIsDir, result);
        }

        STDMETHODIMP CArchiveExtractCallback::GetStream(UInt32 index,
                                                        ISequentialOutStream **outStream,
                                                        Int32 askExtractMode) {
            if (_outFileStream != NULL)
                _outFileStream.Release();
            _outFileStream = NULL;

            WCHAR fullPath[1024];
            fullPath[0] = 0;
            {
                // Get Name
                NCOM::CPropVariant prop;
                RINOK(_archiveHandler->GetProperty(index, kpidPath, &prop));

                if (prop.vt == VT_EMPTY) {
                    WCHAR *pname = mywcsrchr(_archiveBaseName, '/');
                    if (pname)
                        pname++;
                    else
                        pname = _archiveBaseName;
                    if (mywcsncmp(_archiveBaseName,L"content:",8)==0) {
                        WCHAR fileName2[1024];
                        decodeUrl(fileName2,pname,1024);
                        pname = mywcsrchr(fileName2, '/');
                        if (pname)
                            pname++;
                        else
                            pname = fileName2;
                        mywcslcpy(fullPath, pname, 1023);
                    } else
                        mywcslcpy(fullPath, pname, 1023);
                    pname = mywcsrchr(fullPath, '.');
                    if (index != 0 || !pname || pname == fullPath)
                        mywcscpy(fullPath, L"noname");
                    else if (_wcsicmp(pname, L".t7z") == 0 || _wcsicmp(pname, L".tgz") == 0 ||
                             _wcsicmp(pname, L".txz") == 0 ||
                             _wcsicmp(pname, L".tbz") == 0 || _wcsicmp(pname, L".tbz2") == 0 ||
                             _wcsicmp(pname, L".tz") == 0)
                        mywcslcpy(pname, L".tar", 4);
                    else
                        pname[0] = 0;
                } else {
                    if (prop.vt != VT_BSTR) {
                        errorReason |= UN7ZIPERROR_DATAERROR;
                        return E_FAIL;
                    }
                    mywcslcpy(fullPath, prop.bstrVal, 1023);
                }
            }
            //for (size_t i=0;i<mywcslen(fullPath);i++)
            //    __android_log_print(ANDROID_LOG_VERBOSE, "getStream", "wchar=%c", fullPath[i]);

            if (askExtractMode != NArchive::NExtract::NAskMode::kExtract)
                return S_OK;

            {
                // Get Attrib
                NCOM::CPropVariant prop;
                RINOK(_archiveHandler->GetProperty(index, kpidAttrib, &prop));
                if (prop.vt == VT_EMPTY) {
                    _processedFileInfo.Attrib = 0;
                    _processedFileInfo.AttribDefined = false;
                } else {
                    if (prop.vt == VT_UI4) {
                        _processedFileInfo.Attrib = prop.ulVal;
                        _processedFileInfo.AttribDefined = true;
                    }
                }
            }

            RINOK(IsArchiveItemFolder(_archiveHandler, index, _processedFileInfo.isDir));

            {
                // Get Modified Time
                NCOM::CPropVariant prop;
                RINOK(_archiveHandler->GetProperty(index, kpidMTime, &prop));
                _processedFileInfo.MTimeDefined = false;
                switch (prop.vt) {
                    case VT_EMPTY:
                        // _processedFileInfo.MTime = _utcMTimeDefault;
                        break;
                    case VT_FILETIME:
                        _processedFileInfo.MTime = prop.filetime;
                        _processedFileInfo.MTimeDefined = true;
                        break;
                }

            }
            UInt64 newFileSize = -1;
            {
                // Get Size
                NCOM::CPropVariant prop;
                RINOK(_archiveHandler->GetProperty(index, kpidSize, &prop));
                bool newFileSizeDefined = (prop.vt != VT_EMPTY);
                if (newFileSizeDefined)
                    newFileSize = ConvertPropVariantToUInt64(prop);
                else
                    newFileSize = -1;
            }

            WCHAR *pname = fullPath;
            if (subStart < 0) {   // do not keep paths
                pname = mywcsrchr(fullPath, '/');
                if (pname == NULL)
                    pname = fullPath;
            } else if (subStart > 0) {  // extract from subdir
                if (subStart < (int) mywcslen(fullPath)) {
                    pname = fullPath + subStart;
                    if (pname[0] == '/')
                        pname++;
                }
            }

            mywcslcpy(_diskFilePath, _directoryPath, 1023);
            mywcslcat(_diskFilePath, pname, 1023);
            _lastopeneditem = index;

            int extract = EXTRACT_ABORT;

            JNIEnv *envLocal;
            int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);

            jmethodID onWantExtractLocal;
            jclass callbackClass = NULL;
            if (status == JNI_EDETACHED) {
                status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                callbackClass = envLocal->GetObjectClass(callback);
                onWantExtractLocal = envLocal->GetMethodID(callbackClass, "onWantExtract",
                                                           "(ILjava/lang/String;Ljava/lang/String;ZZJJ)Ljava/lang/String;");
            } else {
                status = -12345;
                onWantExtractLocal = onWantExtract;
            }

            char trgbuf[1024];
            WCHAR *src = fullPath;
            char *trg = trgbuf;
            ConvertUCS4toUTF8(&src, fullPath + mywcslen(fullPath) + 1, &trg,
                              trgbuf + sizeof(trgbuf));

            jstring jfullPath = envLocal->NewStringUTF(GetAnsiString(fullPath));
            jstring jdiskFilePath = envLocal->NewStringUTF(GetAnsiString(_diskFilePath));

            // this must also check for presence of file!
            jobject extractTrg = envLocal->CallObjectMethod(callback, onWantExtractLocal, index,
                                                            jfullPath, jdiskFilePath,
                                                            _processedFileInfo.isDir, false,
                                                            newFileSize,
                                                            _processedFileInfo.MTime);
            if (extractTrg != 0) {
                const char *pwstr = envLocal->GetStringUTFChars((jstring) extractTrg, 0);
                if (pwstr) {
                    if (pwstr[0] == 0)
                        extract = EXTRACT_SKIP;
                    else
                        extract = EXTRACT_EXTRACT;
                    mywcslcpy(_diskFilePath, GetUnicodeString(pwstr), 1023);
                    envLocal->ReleaseStringUTFChars((jstring) extractTrg, pwstr);

                }
                envLocal->DeleteLocalRef(extractTrg);
            } else
                extract = EXTRACT_ABORT;

            envLocal->DeleteLocalRef(jfullPath);
            envLocal->DeleteLocalRef(jdiskFilePath);

            if (extract == EXTRACT_SKIP) {
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
                return S_OK;  // SKIP
            } else if (extract == EXTRACT_ABORT) {
                errorReason |= UN7ZIPERROR_ABORTED;
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
                return E_ABORT;
            }

            __android_log_write(ANDROID_LOG_VERBOSE, "Un7zip", trgbuf);

            {
                // Create folders for file
                UString _filePath = _diskFilePath;
                int slashPos = _filePath.ReverseFind(WCHAR_PATH_SEPARATOR);
                if (slashPos >= 0)
                    NFile::NDir::CreateComplexDir(_filePath.Left(slashPos));
            }

            if (_processedFileInfo.isDir) {
                if (!NFile::NDir::CreateComplexDir(_diskFilePath)) {
                    jmethodID onCreateDirLocal;
                    if (callbackClass != NULL) {  // already set above!
                        onCreateDirLocal = envLocal->GetMethodID(callbackClass, "onCreateDir",
                                                                 "(Ljava/lang/String;)I");
                    } else {
                        onCreateDirLocal = onCreateDir;
                    }
                    jstring jfullPath = envLocal->NewStringUTF(GetAnsiString(_diskFilePath));
                    envLocal->CallIntMethod(callback, onCreateDirLocal, jfullPath);
                    envLocal->DeleteLocalRef(jfullPath);
                }
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
            } else {
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
                if (extract == EXTRACT_TOMEM) {
                    WCHAR *p = mywcschr(_diskFilePath, 9);
                    int bufsize = 1024 * 1024;
                    if (p) {
                        p[0] = 0;
                        bufsize = atoiw(p + 1);
                    }
                    long bufaddr = atoilw(_diskFilePath + 1);
                    _outMemStreamSpec = new COutFixedMemStream;
                    CMyComPtr<ISequentialOutStream> outStreamLoc(_outMemStreamSpec);
                    if (!_outMemStreamSpec->Create((char *) bufaddr, bufsize, NULL)) {
                        errorReason |= UN7ZIPERROR_CREATE;
                        return E_ABORT;
                    }
                    _outFileStream = outStreamLoc;
                    *outStream = outStreamLoc.Detach();
                } else if (_testOnly) {
                    _outTestOnlyStreamSpec = new COutTestOnlyStream;
                    CMyComPtr<ISequentialOutStream> outStreamLoc(_outTestOnlyStreamSpec);
                    if (!_outTestOnlyStreamSpec->Open()) {
                        errorReason |= UN7ZIPERROR_CREATE;
                        return E_ABORT;
                    }
                    _outFileStream = outStreamLoc;
                    *outStream = outStreamLoc.Detach();
                } else {

                    NFile::NFind::CFileInfo fi;
                    if (fi.Find(_diskFilePath)) {
                        NFile::NDir::DeleteFileAlways(_diskFilePath);
                        // Ghisler: Ignore error, we may not be allowed to delete here!
                    }

                    _outFileStreamSpec = new COutFileStream;
                    CMyComPtr<ISequentialOutStream> outStreamLoc(_outFileStreamSpec);
                    if (!_outFileStreamSpec->Open(_diskFilePath, CREATE_ALWAYS)) {
                        _outFileStreamSpec = NULL;
                        _outDocStreamSpec = new COutDocStream;
                        CMyComPtr<ISequentialOutStream> outStreamLoc2(_outDocStreamSpec);
                        if (!_outDocStreamSpec->Open(_diskFilePath)) {
                            errorReason |= UN7ZIPERROR_CREATE;
                            return E_ABORT;
                        }
                        _outFileStream = outStreamLoc2;
                        *outStream = outStreamLoc2.Detach();
                    } else {
                        _outFileStream = outStreamLoc;
                        *outStream = outStreamLoc.Detach();
                    }
                }
            }
            return S_OK;
        }

        STDMETHODIMP CArchiveExtractCallback::PrepareOperation(Int32 askExtractMode) {
            _extractMode = false;
            switch (askExtractMode) {
                case NArchive::NExtract::NAskMode::kExtract:
                    _extractMode = true;
                    break;
            };
            switch (askExtractMode) {
                case NArchive::NExtract::NAskMode::kExtract:
                    break;
                case NArchive::NExtract::NAskMode::kTest:
                    break;
                case NArchive::NExtract::NAskMode::kSkip:
                    break;
            };
            return S_OK;
        }

        STDMETHODIMP CArchiveExtractCallback::SetOperationResult(Int32 operationResult) {
            bool haserror = operationResult != NArchive::NExtract::NOperationResult::kOK;
            if (haserror) {
                NumErrors++;
                switch (operationResult) {
                    case NArchive::NExtract::NOperationResult::kUnsupportedMethod:
                        errorReason |= UN7ZIPERROR_UNSUPPORTED;
                        break;
                    case NArchive::NExtract::NOperationResult::kCRCError:
                        errorReason |= UN7ZIPERROR_CRCERROR;
                        break;
                    case NArchive::NExtract::NOperationResult::kDataError:
                        errorReason |= UN7ZIPERROR_DATAERROR;
                        break;
                    default:
                        //PrintString(kUnknownError);
                        ;
                }
            }

            if (_outFileStream != NULL && _outFileStreamSpec != NULL) {
                if (_processedFileInfo.MTimeDefined && !haserror)
                    _outFileStreamSpec->SetMTime(&_processedFileInfo.MTime);
                RINOK(_outFileStreamSpec->Close());
            } else if (_outFileStream != NULL && _outDocStreamSpec != NULL) {
                RINOK(_outDocStreamSpec->Close());
            }
            if (_outFileStream != NULL) {
                _outFileStream.Release();
                _outFileStream = NULL;
            }
            //if (_extractMode && _processedFileInfo.AttribDefined && !haserror)
//        NFile::NDirectory::MySetFileAttributes(_diskFilePath, _processedFileInfo.Attrib);
            if (haserror)
                return E_FAIL;
            _lastsuccessfullycloseditem = _lastopeneditem;
            return S_OK;
        }

        STDMETHODIMP CArchiveExtractCallback::CryptoGetTextPassword(BSTR *password) {
            if (onGetPassword != 0) {
                JNIEnv *envLocal;
                int status = gJavaVM->GetEnv((void **) &envLocal, JNI_VERSION_1_6);
                jmethodID onGetPasswordLocal;
                if (status == JNI_EDETACHED) {   // called from different thread?
                    status = gJavaVM->AttachCurrentThread(&envLocal, NULL);
                    jclass callbackClass = envLocal->GetObjectClass(callback);
                    onGetPasswordLocal = envLocal->GetMethodID(callbackClass, "onGetPassword",
                                                               "()Ljava/lang/String;");
                } else {
                    status = -12345;
                    onGetPasswordLocal = onGetPassword;
                }
                jobject pw = envLocal->CallObjectMethod(callback, onGetPasswordLocal);
                if (pw != 0) {
                    const char *pwstr = envLocal->GetStringUTFChars((jstring) pw, 0);
                    if (pwstr != NULL) {
                        int ret = StringToBstr(GetUnicodeString(pwstr), password);
                        envLocal->ReleaseStringUTFChars((jstring) pw, pwstr);
                        if (status == 0)
                            gJavaVM->DetachCurrentThread();
                        return S_OK;
                    }
                }
                if (status == 0)
                    gJavaVM->DetachCurrentThread();
            }
            /*  if (PasswordCallback) {
                  WCHAR buf[256];
                  buf[0] = 0;
                  if (PasswordCallback(buf, COUNTOF(buf) - 1, _userData)) {
                      return StringToBstr(GetUnicodeString(buf), password);
                  }
              }*/
            return E_ABORT;
        }


#ifdef __cplusplus
        extern "C" {
#endif

/*JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM* vm, void* reserved) {
    gJavaVM = vm;
    return JNI_VERSION_1_6;
}
*/

        const GUID *getGuidForName(const char *name, BOOL *isRAR) {
            *isRAR = false;
            char ext[64];
            const char *pext = strrchr(name, '.');
            if (pext == NULL)
                return &CLSID_CFormat7z2;
            strlcpy(ext, pext + 1, sizeof(ext) - 1);
            if (strcasecmp(ext, "tar") == 0)
                return &CLSID_CFormattar;
            if (strcasecmp(ext, "arj") == 0)
                return &CLSID_CFormatarj;
            if (strcasecmp(ext, "rar") == 0) {
                *isRAR = true;
                return &CLSID_CFormatrar;
            }
            if (strcasecmp(ext, "lzh") == 0 || strcasecmp(ext, "lha") == 0)
                return &CLSID_CFormatlzh;
            if (strcasecmp(ext, "bz2") == 0 || strcasecmp(ext, "bzip2") == 0 ||
                strcasecmp(ext, "tbz") == 0 || strcasecmp(ext, "tbz2") == 0)
                return &CLSID_CFormatbzip2;
            if (strcasecmp(ext, "gz") == 0 || strcasecmp(ext, "gzip") == 0 ||
                strcasecmp(ext, "tgz") == 0)
                return &CLSID_CFormatgz;
//    if (strcasecmp(ext,"xz")==0 || strcasecmp(ext,"txz")==0)
//        return &CLSID_CFormatxz;
            if (strcasecmp(ext, "z") == 0 || strcasecmp(ext, "tz") == 0)
                return &CLSID_CFormatz;
            if (strcasecmp(ext, "cab") == 0)
                return &CLSID_CFormatcab;
            if (strcasecmp(ext, "iso") == 0 || strcasecmp(ext, "img") == 0)
                return &CLSID_CFormatiso;
            return &CLSID_CFormat7z2;
        }

        void loadStreamCallbackFunctions(JNIEnv *env, jobject streamCallback) {
            if (gJavaVM == NULL)
                env->GetJavaVM(&gJavaVM);
            if (globalStreamCallback != NULL)
                env->DeleteGlobalRef(globalStreamCallback);
            globalStreamCallback = env->NewGlobalRef(
                    streamCallback);   //MUST be preserved, because this function can return before the last file is closed!
            jclass streamCallbackClass = env->GetObjectClass(globalStreamCallback);
            openTargetDocumentGlobal = env->GetMethodID(streamCallbackClass, "openTargetDocument",
                                                        "(Ljava/lang/String;)I");
            openSourceDocumentGlobal = env->GetMethodID(streamCallbackClass, "openSourceDocument",
                                                        "(Ljava/lang/String;)I");
            getFdGlobal = env->GetMethodID(streamCallbackClass, "getFd", "(I)I");
            closeDocumentGlobal = env->GetMethodID(streamCallbackClass, "closeDocument", "(I)V");
            writeTargetDocumentGlobal = env->GetMethodID(streamCallbackClass, "writeTargetDocument",
                                                         "(I[BI)I");
        }

/************************************************************************************/
/******************************** Exported functions ********************************/
/************************************************************************************/

        JNIEXPORT jint JNICALL
        Java_org_ghisler_a7zip_Un7zipjniLib_extract(JNIEnv *env, jobject obj,
                                                                     jstring name,
                                                                     jint subStart,
                                                                     jstring destination,
                                                                     jobject callback,
                                                                     jobject streamCallback) {
            loadStreamCallbackFunctions(env, streamCallback);
            const char *nativeName = env->GetStringUTFChars(name, 0);
            const char *nativeDestination = env->GetStringUTFChars(destination, 0);

            CMyComPtr<IInArchive> archive = NULL;
            BOOL isRAR = false;
            if (CreateArchiver(getGuidForName(nativeName, &isRAR), &IID_IInArchive2,
                               (void **) &archive) != S_OK) {
                env->ReleaseStringUTFChars(name, nativeName);
                env->ReleaseStringUTFChars(destination, nativeDestination);
                return -2;
            }
            CMyInFileStream *fileSpec = new CMyInFileStream;
            CMyComPtr<IInStream> file = fileSpec;

            if (!fileSpec->Open(GetUnicodeString(nativeName))) {
                __android_log_write(ANDROID_LOG_VERBOSE, "un7zip", "fileSpec->Open failed!");
                env->ReleaseStringUTFChars(name, nativeName);
                env->ReleaseStringUTFChars(destination, nativeDestination);
                return -3;
            }
            const char *p = strrchr(nativeName, '.');
            if (p && p[1] == '0' && p[2] == '0' && p[3] == '1') {
                fileSpec->VolumeBaseName[0] = 0;
                mywcslcat(fileSpec->VolumeBaseName, GetUnicodeString(nativeName), 1023);
                fileSpec->Seek(0, 2, &fileSpec->FirstVolumeSize);
                fileSpec->CurrentVolumeNr = 1;
                WCHAR *p2 = mywcsrchr(fileSpec->VolumeBaseName, '.');
                if (p2)
                    p2[1] = 0;
            }
            {
                CArchiveOpenCallback *openCallbackSpec = new CArchiveOpenCallback;
                openCallbackSpec->Init(env, callback,GetUnicodeString(nativeName)); // second parameter is output folder path
                CMyComPtr<IArchiveOpenCallback> openCallback(openCallbackSpec);
//        openCallbackSpec->_userData=userData;

                if (archive->Open(file, 0, openCallback) != S_OK) {
                    if (isRAR) {  // try RAR5!
                        if (CreateArchiver(&CLSID_CFormatrar5, &IID_IInArchive2,
                                           (void **) &archive) != S_OK) {
                            env->ReleaseStringUTFChars(name, nativeName);
                            env->ReleaseStringUTFChars(destination, nativeDestination);
                            return -2;
                        }
                        fileSpec->Seek(0, SEEK_SET, NULL);
                        if (archive->Open(file, 0, openCallback) != S_OK) {
                            __android_log_write(ANDROID_LOG_VERBOSE, "un7zip",
                                                "archive->Open failed2!");
                            archive = NULL;
                            env->ReleaseStringUTFChars(name, nativeName);
                            env->ReleaseStringUTFChars(destination, nativeDestination);
                            return -4;
                        }
                    } else {
                        __android_log_write(ANDROID_LOG_VERBOSE, "un7zip",
                                            "archive->Open failed1!");
                        archive = NULL;
                        env->ReleaseStringUTFChars(name, nativeName);
                        env->ReleaseStringUTFChars(destination, nativeDestination);
                        return -4;
                    }
                }
            }

            UInt32 numItems;
            archive->GetNumberOfItems(&numItems);

            openFileCount = 0;
            jobject globalCallback = env->NewGlobalRef(callback);
            jclass callbackClass = env->GetObjectClass(globalCallback);
            jmethodID onGetFileCount = env->GetMethodID(callbackClass, "onGetFileCount", "(I)V");
            if (onGetFileCount != 0)
                env->CallVoidMethod(callback, onGetFileCount, numItems);

            __android_log_print(ANDROID_LOG_VERBOSE, "Un7zip", "Item count: %d", numItems);

            CArchiveExtractCallback *extractCallbackSpec = new CArchiveExtractCallback;
            CMyComPtr<IArchiveExtractCallback> extractCallback(extractCallbackSpec);

            extractCallbackSpec->Init(archive, GetUnicodeString(nativeDestination),
                                      strncmp(nativeDestination, "/dev/null", 9) == 0, env,
                                      globalCallback); // second parameter is output folder path
            extractCallbackSpec->subStart = subStart;
            extractCallbackSpec->errorReason = 0;
            extractCallbackSpec->_lastsuccessfullycloseditem = 0xFFFFFFFF;
            extractCallbackSpec->_lastopeneditem = 0xFFFFFFFF;
            mywcslcpy(extractCallbackSpec->_archiveBaseName, GetUnicodeString(nativeName), 1023);

            HRESULT ret2 = archive->Extract(NULL, 0xFFFFFFFF,
                                                strncmp(nativeDestination, "/dev/null", 9) == 0,
                                                extractCallback);
            env->DeleteGlobalRef(globalCallback);
            env->ReleaseStringUTFChars(name, nativeName);
            env->ReleaseStringUTFChars(destination, nativeDestination);
            archive->Close();
            globalCallback = NULL;
            if (ret2 == 0x8007000E)
                extractCallbackSpec->errorReason |= UN7ZIPERROR_MEMORY;
            if (ret2 == 0)
                return 0;
            else
                return extractCallbackSpec->errorReason;
        }

        JNIEXPORT jint JNICALL
        Java_org_ghisler_a7zip_Un7zipjniLib_list(JNIEnv *env, jobject obj,
                                                                  jstring name,
                                                                  jobject callback,
                                                                  jobject streamCallback) {
            loadStreamCallbackFunctions(env, streamCallback);
            const char *nativeName = env->GetStringUTFChars(name, 0);

            CMyComPtr<IInArchive> archive = NULL;
            BOOL isRAR = false;
            if (CreateArchiver(getGuidForName(nativeName, &isRAR), &IID_IInArchive2,
                               (void **) &archive) != S_OK) {
                env->ReleaseStringUTFChars(name, nativeName);
                return UN7ZIPERROR_UNSUPPORTED;
            }
            CMyInFileStream *fileSpec = new CMyInFileStream;
            CMyComPtr<IInStream> file = fileSpec;

            if (!fileSpec->Open(GetUnicodeString(nativeName))) {
                env->ReleaseStringUTFChars(name, nativeName);
                return UN7ZIPERROR_DATAERROR;
            }

            int ret=0;
            // Note: the 7zip handler doesn't call the volume callback! That only works for RAR
            const char *p = strrchr(nativeName, '.');
            if (p && p[1] == '0' && p[2] == '0' && p[3] == '1') {
                fileSpec->VolumeBaseName[0] = 0;
                mywcslcat(fileSpec->VolumeBaseName, GetUnicodeString(nativeName), 1023);
                fileSpec->Seek(0, 2, &fileSpec->FirstVolumeSize);
                fileSpec->CurrentVolumeNr = 1;
                WCHAR *p2 = mywcsrchr(fileSpec->VolumeBaseName, '.');
                if (p2)
                    p2[1] = 0;
            }

            {
                CArchiveOpenCallback *openCallbackSpec = new CArchiveOpenCallback;
                openCallbackSpec->Init(env, callback,GetUnicodeString(nativeName)); // second parameter is output folder path
                CMyComPtr<IArchiveOpenCallback> openCallback(openCallbackSpec);

                if (archive->Open(file, 0, openCallback) != S_OK) {
                    if (isRAR) {  // try RAR5!
                        if (CreateArchiver(&CLSID_CFormatrar5, &IID_IInArchive2,
                                           (void **) &archive) != S_OK) {
                            env->ReleaseStringUTFChars(name, nativeName);
                            return UN7ZIPERROR_UNSUPPORTED;
                        }
                        fileSpec->Seek(0, SEEK_SET, NULL);
                        if (archive->Open(file, 0, openCallback) != S_OK) {
                            archive = NULL;
                            env->ReleaseStringUTFChars(name, nativeName);
                            return UN7ZIPERROR_DATAERROR;
                        }
                    } else {
                        archive = NULL;
                        env->ReleaseStringUTFChars(name, nativeName);
                        return UN7ZIPERROR_DATAERROR;
                    }
                }
            }

            UInt32 numItems;
            archive->GetNumberOfItems(&numItems);

            jclass callbackClass = env->GetObjectClass(callback);
            jmethodID onGetFileCount = env->GetMethodID(callbackClass, "onGetFileCount", "(I)V");
            jmethodID onListItem = env->GetMethodID(callbackClass, "onListItem",
                                                    "(ILjava/lang/String;ZZJJ)I");
            if (onGetFileCount != 0)
                env->CallVoidMethod(callback, onGetFileCount, numItems);
            __android_log_print(ANDROID_LOG_VERBOSE, "Un7zip", "Item count: %d", numItems);

            for (int i = 0; i < numItems; i++) {
                // Get name of file
                BOOL ok = false;
                WCHAR fileName[1024];
                NWindows::NCOM::CPropVariant prop;
                archive->GetProperty(i, kpidPath, &prop);
                fileName[0] = 0;
                if (prop.vt == VT_BSTR) {
                    mywcslcat(fileName, prop.bstrVal, 1023);
                    fileName[COUNTOF(fileName) - 1] = 0;
                    ok = true;
                } else if (prop.vt == 0) {  // an empty name -> use archive name without extension
                    WCHAR *pname = mywcsrchr(fileSpec->VolumeBaseName, '/');
                    if (pname)
                        pname++;
                    else
                        pname = fileSpec->VolumeBaseName;

                    if (mywcsncmp(fileSpec->VolumeBaseName,L"content:",8)==0) {
                        WCHAR fileName2[1024];
                        decodeUrl(fileName2,pname,1024);
                        pname = mywcsrchr(fileName2, '/');
                        if (pname)
                            pname++;
                        else
                            pname = fileName2;
                        mywcslcpy(fileName, pname, 1023);
                    } else
                        mywcslcpy(fileName, pname, 1023);

                    pname = mywcsrchr(fileName, '.');
                    if (i != 0 || !pname || pname == fileName)
                        mywcscpy(fileName, L"noname");
                    else if (_wcsicmp(pname, L".t7z") == 0 || _wcsicmp(pname, L".tgz") == 0 ||
                             _wcsicmp(pname, L".txz") == 0 ||
                             _wcsicmp(pname, L".tbz") == 0 || _wcsicmp(pname, L".tbz2") == 0 ||
                             _wcsicmp(pname, L".tz") == 0)
                        mywcslcpy(pname, L".tar", 4);
                    else
                        pname[0] = 0;
                    ok = true;
                }
                if (ok) {
                    jlong size = 0;
                    jlong time = 0;
                    if (archive->GetProperty(i, kpidSize, &prop) == 0)
                        size = ConvertPropVariantToUInt64(prop);
                    if (archive->GetProperty(i, kpidMTime, &prop) == 0) {
                        if (prop.vt == VT_FILETIME) {
                            time = prop.filetime.dwHighDateTime;
                            time = time << 32 | prop.filetime.dwLowDateTime;
                        }
                    }
                    BOOL isDir = false;
                    if (archive->GetProperty(i, kpidAttrib, &prop) == 0) {
                        isDir = (ConvertPropVariantToUInt64(prop) & FILE_ATTRIBUTE_DIRECTORY) != 0;
                    }
                    if (archive->GetProperty(i, kpidIsDir, &prop) == 0) {
                        if (ConvertPropVariantToUInt64(prop) != 0)
                            isDir = true;
                    }
                    BOOL isEncrypted = false;
                    if (archive->GetProperty(i, kpidEncrypted, &prop) == 0) {
                        isEncrypted = (ConvertPropVariantToUInt64(prop) != 0);
                    }
                    //__android_log_print(ANDROID_LOG_VERBOSE, "Un7zip","Adding item: %d",i);
                    AString astr = GetAnsiString(fileName);
                    char *buf = astr.GetBuf(0);
                    jstring jfullPath = env->NewStringUTF(buf);
                    ret=env->CallIntMethod(callback, onListItem, i, jfullPath,
                                        isDir, isEncrypted, size, time);
                    env->DeleteLocalRef(jfullPath);
                    if (ret!=0) {
                        ret=UN7ZIPERROR_ABORTED;
                        break;
                    }
                }
            }

            env->ReleaseStringUTFChars(name, nativeName);
            archive->Close();
            return ret;
        }

#ifdef __cplusplus
        }
#endif
    }
}
